-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

Vehicle = extends (ColClass) {

        castShadows = true;
        renderingDistance=100;

        steerMax = 25;
        steerRate = 1.3;
        unsteerRate = 5;
        steerFastSpeed = 100;
        steerRateFast = 0.3;
        unsteerRateFast = 5;

        parkedDampingFactor = 0.9;

        glideDirection = Vector3(0,1,0);
        glideFactor = -6;
        glideDampingFactor = 2;

        wheelSpinTractionControlMax = 1;
        wheelSpinTractionControlRate = 4;
        wheelSpinTractionControlMin = 0.01;

        smokeSystemArg = { longevity=2, grow=2, startAlpha=1, period=20 };

        brakePlot = { [-10] = 100; [0] = 100; [10] = 2500; [20] = 2500; [40] = 2500; };

}


function Vehicle.activate (persistent,class,gfxRoot,physics)
        local self = ColClass.activate(persistent,class,gfxRoot,physics)

        self.smokeSystem = sm:createParticleSystem(persistent.name..":ParticleSystem")
        self.gfx:attachObject(self.smokeSystem)
        self.smokeSystem.materialName = "common/Particle/Smoke"
        self.smokeTimer = 0

        self.handbrake = false
        self.brake = false
        self.push = 0
        self.parked = true

        self.driftHackTimer = 0

        local mpw = self.body.mass/#class.wheelInfo -- mass per wheel (hacky)
        local len = class.suspensionLength

        self.steer, self.shouldSteerLeft, self.shouldSteerRight = 0, 0, 0 
        self.tractionControl = 1
        local n = self.gfx

        self.engine = Engine.new(class.powerPlots)

        self.brakeCurve = Plot (class.brakePlot)

        self.wheels = {}
        self.drivenWheels = {}
        self.handbrakeWheels = {}

        for i, info in ipairs(class.wheelInfo) do
                local wmass = (info.mass or 1) * mpw
                self.wheels[i] = WheelAssembly.new(info.left, self, self.gfx, fqn_ex(info.mesh, class.name):sub(2),
                                                   info.pos, info.rad, info.len, info.steer,
                                                   wmass, info.mu, info.down,
                                                   info.forwards, info.dampingFactor, info.offRoad)
                self.wheels[i].drive = info.drive
                self.wheels[i].wheelEntity.castShadows = class.castShadows

                if info.drive then table.insert(self.drivenWheels,self.wheels[i]) end
                if info.handbrake then table.insert(self.handbrakeWheels,self.wheels[i]) end
        end 

        self.body.updateCallback = function (px,py,pz,rw,rx,ry,rz)
                local node = self.gfx
                -- can be nil
                if node ~= nil and alive(node) then
                        node:setPosition(px,py,pz)
                        node:setOrientation(rw,rx,ry,rz)
                end
                foreach (self.wheels, WheelAssembly.updateGFX)
                persistent:updateSphere(px,py,pz,self.renderingDistance)
        end

        self.body.stepCallback = function()

                local body = self.body
                local world = body.world
                local velocity = body.linearVelocity
                local speed = #velocity

                -- hack to stop parked cars drifting
                if self.parked or self.brake or self.handBrake then
                        -- if we're parked then soak up anything left over from last time
                        if self.driftHackTimer > 100 then
                                if speed<0.04 then
                                        body.linearVelocity = V_ZERO
                                        body.angularVelocity = V_ZERO
                                end
                        end
                end

                local forward_speed = velocity:dot(body.worldOrientation * V_FORWARDS)
                local mph = forward_speed * 60 * 60 / METRES_PER_MILE

                local steerTarget = self.shouldSteerRight + self.shouldSteerLeft
                local change = steerTarget - self.steer
                if math.abs(change) > 0.001 then
                        local rate
                        local fastness = clamp(mph/class.steerFastSpeed, 0, 1)
                        if between(0,steerTarget,self.steer) and math.abs(self.steer) > 0.001 then
                                rate = (1-fastness)*class.unsteerRate + (fastness)*class.unsteerRateFast
                        else
                                rate = (1-fastness)*class.steerRate + (fastness)*class.steerRateFast
                        end
                        change = clamp(change,-rate, rate)
                        self.steer = self.steer + change
                        for _,w in ipairs(self.wheels) do
                                w:setSteer(self.steer)
                        end
                end
                
                local torque = self.engine:getTorque(mph)
        
                for _, wheel in ipairs(self.wheels) do
                        wheel.locked = self.parked
                        wheel:setTorque(0)
                end
                local tcstep = class.wheelSpinTractionControlRate * world.stepSize
                if self.handbrake and self.engine:getGear() ~= 0 then
                        -- 'wheelspin'
                        self.tractionControl = self.tractionControl - tcstep
                else
                        self.tractionControl = self.tractionControl + tcstep
                end
                self.tractionControl = clamp(self.tractionControl,
                                             class.wheelSpinTractionControlMin,
                                             class.wheelSpinTractionControlMax)
                for _, wheel in ipairs(self.handbrakeWheels) do
                        if self.handbrake then
                                if self.engine:getGear() == 0 or not wheel.drive then
                                        -- handbrake turn
                                        wheel.locked = true
                                end
                        end
                        if wheel.drive then
                                wheel.tractionControl = self.tractionControl
                        end
                end
                for _, wheel in ipairs(self.drivenWheels) do
                        wheel:setTorque(wheel.drive * torque / #self.drivenWheels)
                end
                local arg = class.smokeSystemArg
                self.smokeTimer = self.smokeTimer + 1
                local emit_smoke = false
                if self.smokeTimer >= arg.period / math.max(speed*0.3,1) then
                        self.smokeTimer = 0
                        emit_smoke = true
                end
                for _, wheel in ipairs(self.wheels) do
                        if self.brake then
                                if self.handbrake then
                                        wheel.locked = true
                                else
                                        wheel:setTorque(-sign(mph)*self.brakeCurve[math.abs(mph)])
                                end
                        end
                        wheel:process()
                        local rnd = math.random
                        local skidding = wheel.skidding
                        if not wheel.locked then
                                -- must be skidding due to wheelspin
                                -- so attenuate smoke based on wheelspin
                                skidding = skidding - invlerp(self.tractionControl,
                                                              class.wheelSpinTractionControlMin,
                                                              class.wheelSpinTractionControlMax)
                        end
                        
                        if skidding > 0.0 then
                                if emit_smoke then
                                        local px, py, pz = wheel.worldContactPos:xyz()
                                        self.smokeSystem:addParticle(px,py,pz+0.25, rnd()*360, --pos,rot
                                                                     0.5+rnd(),0.5+rnd(), --scale
                                                                     1,1,1, --colour
                                                                     skidding*arg.startAlpha*(1-rnd()/3),
                                                                     0.2-0.4*rnd(), 0.2-0.4*rnd(), 0.6+rnd(),
                                                                     particle_smoke, arg, 0)
                                end
                        end
                end

        end

        self.body.stabiliseCallback = function ()

                local body = self.body
                local world = self.world

                -- tend the travel direction towards the glide direction
                local travelDirection = body.linearVelocity.normalised -- :randomDeviant(0).normalised
                local spin = body.angularVelocity
                local sheer = travelDirection:cross(body.worldOrientation * class.glideDirection)
                local torque = class.glideFactor * sheer - class.glideDampingFactor * spin
                local speed = math.min(#body.linearVelocity, 125)
                body:torque(speed * speed * torque)
                
                -- hack to stop parked cars drifting
                if self.parked or self.brake or self.handBrake then
                        local v = body.linearVelocity
                        local gdir = world.gravity.normalised
                        -- derive the anti-gravity component of v
                        local vz = v:dot(gdir) * gdir
                        -- work out what our velocity will be after gravity is applied
                        local final_v = v + world.gravity*world.stepSize
                        -- if the resultant force of this frame was small (i.e. was jitter)
                        if #final_v<0.1 then
                                self.driftHackTimer = self.driftHackTimer + 1
                                if self.driftHackTimer > 100 then
                                        -- soak up this force (except for anti-gravity component)
                                        body.linearVelocity = vz
                                        body.angularVelocity = V_ZERO
                                end
                        end
                else
                        self.driftHackTimer = 0
                end
        end
        return self
end

function Vehicle:frameCallback (time)
        self.instance.smokeSystem:update(time)
end

function Vehicle:deactivate (instance)
        foreach(instance.wheels, safe_destroy)
        safe_destroy(instance.smokeSystem)
        instance.wheels = {}
        instance.drivenWheels = {}
        instance.handbrakeWheels = {}
        return ColClass.deactivate(self, instance)
end

function Vehicle:setFade (instance, fade, transition)
        for _, wheel in ipairs(instance.wheels) do
                local ent = wheel.wheelEntity
                if transition then
                        -- slow path
                        transition_entity(ent, transition)
                end
                ent:setCustomParameterAll(0,fade,0,0,0)
        end
        ColClass.setFade(self, instance, fade, transition)
end


function Vehicle:getSpeed ()
        if not self.activated then error("Not activated: "..self.persistent.name) end
        local rb = self.instance.body
        return #rb.linearVelocity
end

function Vehicle:setSpeed (speed)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        local rb = self.instance.body
        rb.linearVelocity = rb.linearVelocity.normalised * speed
        rb:activate();
end

function Vehicle:boost (speed)
        speed = speed or 1
        self:setSpeed(self:getSpeed()+speed)
end

function Vehicle:setHandbrake (v)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        self.instance.handbrake = v
end

function Vehicle:abandon ()
        if not self.activated then return end
        self = self.instance
        local body = self.body
        self.parked = body.linearVelocity:dot(body.worldOrientation * V_FORWARDS) < 0.5
        self.brake = false
        self.engine.gear = 0
        self.push = 0
        self.handbrake = false
end

function Vehicle:changePush ()
        if not self.activated then error("Not activated: "..self.persistent.name) end
        self = self.instance
        local body = self.body
        self.parked = false -- if we push/pull the vehicle at all then it becomes unparked until abandoned
        if self.push == 0 then
                self.engine:setGear(0)
                self.brake = false
        else
                local speed = body.linearVelocity:dot(body.worldOrientation * V_FORWARDS)
                local wrong_way = math.abs(speed) > 7 and sign(speed) ~= sign(self.push)
                if wrong_way then
                        self.brake = true
                else
                        self.engine:setGear(sign(self.push)) -- one gear for forward, one for reverse
                end
        end
end

function Vehicle:setPull (v)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        local change = v and 1 or -1
        self.instance.push = self.instance.push - change
        self:changePush()
end

function Vehicle:setPush (v)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        local change = v and 1 or -1
        self.instance.push = self.instance.push + change
        self:changePush()
end


-- whether key is pressed
function Vehicle:setShouldSteerLeft (v)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        self.instance.shouldSteerLeft = v and -self.steerMax or 0
end
        
-- whether key is pressed
function Vehicle:setShouldSteerRight (v)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        self.instance.shouldSteerRight = v and self.steerMax or 0
end
        

function Vehicle:warp (pos,orientation)
        pos = pos or player_ctrl:pickPos()+Vector3(0,0,2)
        orientation = orientation or gfx.cam.derivedOrientation
        self:activate()
        if not self.activated then error("Tried to activate but still not activated: "..self.persistent.name) end
        local rb = self.instance.body
        rb.worldPosition = pos
        rb.worldOrientation = orientation
        rb.linearVelocity = V_ZERO
        rb.angularVelocity = V_ZERO
end

function Vehicle:realign ()
        if not self.activated then error("Not activated: "..self.persistent.name) end
        local rb = self.instance.body
        rb.worldOrientation = V_NORTH:getRotationTo(rb.worldOrientation * V_FORWARDS * Vector3(1,1,0));
        rb.worldPosition = rb.worldPosition + Vector3(0,0,3);
        rb.angularVelocity = V_ZERO
        rb.linearVelocity = V_ZERO
        rb:activate()
        self.instance.broken = false
end

function Vehicle:setMu (mu)
        if not self.activated then error("Not activated: "..self.persistent.name) end
        for _,w in ipairs(self.instance.wheels) do
                w.mu = mu
        end
end

function Vehicle:reset ()
        if not self.activated then error("Not activated: "..self.persistent.name) end
        self.pos = self.instance.body.worldPosition
        self.rot = self.instance.body.worldOrientation
        self:deactivate()
        self:activate()
        player_ctrl:drive(self)
end


Bike = extends (Vehicle) {

        steerRate = 0.4;
        unsteerRate = 0.3;
        steerFastSpeed = 40;
        steerRateFast = 0.03;
        unsteerRateFast = 0.4;


        glideFactor = -0.2;
        glideDampingFactor = 0.1;

        stabilisationFactor = 4000;
        stabilisationDampingFactor = 1000;
        canterAngle = 25;

        maxLean = 30;
        fallTorque = 4000;

        brakePlot = { [-10] = 400; [0] = 400; [10] = 400; [20] = 400; [40] = 400; };

}

function Bike.activate (persistent,class,gfxRoot,physics)
        local self = Vehicle.activate(persistent,class,gfxRoot,physics)
        local oldStepCallback = self.body.stepCallback
        self.body.stepCallback = function ()
                oldStepCallback()

                local fwheel = self.wheels[1]
                local rwheel = self.wheels[2]

                if self.broken then return end

                if fwheel.onGround or rwheel.onGround then

                        local body = self.body
                        local world = self.world
                        local q = body.worldOrientation

                        local forwards_ws = q * V_FORWARDS

                        local desired_vector


                        if true or (fwheel.onGround and rwheel.onGround) then
                                local dir = fwheel.worldContactPos-rwheel.worldContactPos
                                local speed = body.linearVelocity:dot(dir.normalised)
                                local turn_radius
                                if self.steer == 0 then
                                        turn_radius = math.huge;
                                else
                                        turn_radius = #dir / (math.rad(self.steer) * math.cos(math.rad(class.canterAngle)))
                                end
                                local desired_lean =
                                        math.deg(math.atan((speed*speed) / (#world.gravity*turn_radius)))
                                --desired_lean = desired_lean / 2
                                desired_lean = clamp(desired_lean,-class.maxLean,class.maxLean)
                                desired_vector = Quat(desired_lean,forwards_ws) * V_UP
                        else
                                local dir = fwheel.pos-rwheel.pos
                                local speed = body.linearVelocity:dot(dir.normalised)
                                local turn_radius = #dir
                                         / (math.rad(self.steer) * math.cos(math.rad(class.canterAngle)))
                                local desired_lean =
                                        math.deg(math.atan((speed*speed) / (#world.gravity*turn_radius)))
                                desired_vector = Quat(desired_lean,forwards_ws) * V_UP
                                desired_vector =  V_UP
                        end

                        --echo(desired_vector)

                         --using the actual friction proved too jittery
--[[
                        desired_vector = Vector3(0,0,0)
                        for _,w in ipairs{self.wheels[1], self.wheels[2]} do
                                if w.onGround then
                                        desired_vector:append(w.restitutionForce)
                                        desired_vector:append(w.frictionForce)
                                end
                        end
]]--


                        desired_vector:normalise()

                        local vector = q*V_UP

                        local sheer = vector:cross(desired_vector)
                        sheer = sheer:dot(forwards_ws) * forwards_ws


                        local spin = body.angularVelocity
                        spin = spin:dot(forwards_ws)
                        spin = spin * forwards_ws

                        local torque = class.stabilisationFactor * sheer
                                     - class.stabilisationDampingFactor * spin

                        if #torque > class.fallTorque then
                                self.broken = true
                                return
                        end

                        body:torque(torque)
                end
        end
        return self
end

-- vim: shiftwidth=8:tabstop=8:expandtab
